# Copyright 2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.restfulclient
#
# lazr.restfulclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.restfulclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.restfulclient.  If not, see <http://www.gnu.org/licenses/>.
"Test client for the lazr.restful example web service."

__metaclass__ = type
__all__ = [
    "CookbookWebServiceClient",
]


try:
    # Python 3.
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from lazr.restfulclient.resource import (
    CollectionWithKeyBasedLookup,
    ServiceRoot,
)


class CookbookSet(CollectionWithKeyBasedLookup):
    """A custom subclass capable of cookbook lookup by cookbook name."""

    def _get_url_from_id(self, id):
        """Transform a cookbook name into the URL to a cookbook resource."""
        return (
            str(self._root._root_uri.ensureSlash())
            + "cookbooks/"
            + quote(str(id))
        )

    collection_of = "cookbook"


class RecipeSet(CollectionWithKeyBasedLookup):
    """A custom subclass capable of recipe lookup by recipe ID."""

    def _get_url_from_id(self, id):
        """Transform a recipe ID into the URL to a recipe resource."""
        return str(self._root._root_uri.ensureSlash()) + "recipes/" + str(id)

    collection_of = "recipe"


class CookbookWebServiceClient(ServiceRoot):

    RESOURCE_TYPE_CLASSES = dict(ServiceRoot.RESOURCE_TYPE_CLASSES)
    RESOURCE_TYPE_CLASSES["recipes"] = RecipeSet
    RESOURCE_TYPE_CLASSES["cookbooks"] = CookbookSet

    DEFAULT_SERVICE_ROOT = "http://cookbooks.dev/"
    DEFAULT_VERSION = "1.0"

    def __init__(
        self,
        service_root=DEFAULT_SERVICE_ROOT,
        version=DEFAULT_VERSION,
        cache=None,
    ):
        super(CookbookWebServiceClient, self).__init__(
            None, service_root, cache=cache, version=version
        )



# Copyright 2009-2018 Canonical Ltd.

# This file is part of lazr.restfulclient.
#
# lazr.restfulclient is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, version 3 of the
# License.
#
# lazr.restfulclient is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.restfulclient. If not, see <http://www.gnu.org/licenses/>.

"""Tests for the OAuth-aware classes."""

__metaclass__ = type


import os
import os.path
import stat
import unittest

from fixtures import MockPatch, TempDir
from testtools import TestCase

from lazr.restfulclient.authorize.oauth import (
    AccessToken,
    Consumer,
    OAuthAuthorizer,
    SystemWideConsumer,
)


class TestConsumer(TestCase):
    def test_data_fields(self):
        consumer = Consumer("key", "secret", "application")
        self.assertEqual(consumer.key, "key")
        self.assertEqual(consumer.secret, "secret")
        self.assertEqual(consumer.application_name, "application")

    def test_default_application_name(self):
        # Application name defaults to None
        consumer = Consumer("key", "secret")
        self.assertEqual(consumer.application_name, None)


class TestAccessToken(TestCase):
    def test_data_fields(self):
        access_token = AccessToken("key", "secret", "context")
        self.assertEqual(access_token.key, "key")
        self.assertEqual(access_token.secret, "secret")
        self.assertEqual(access_token.context, "context")

    def test_default_context(self):
        # Context defaults to None.
        access_token = AccessToken("key", "secret")
        self.assertIsNone(access_token.context)

    def test___str__(self):
        access_token = AccessToken("lock&key", "secret=password")
        self.assertEqual(
            "oauth_token_secret=secret%3Dpassword&oauth_token=lock%26key",
            str(access_token),
        )

    def test_from_string(self):
        access_token = AccessToken.from_string(
            "oauth_token_secret=secret%3Dpassword&oauth_token=lock%26key"
        )
        self.assertEqual(access_token.key, "lock&key")
        self.assertEqual(access_token.secret, "secret=password")


class TestSystemWideConsumer(TestCase):
    def test_useful_distro_name(self):
        # If distro.name returns a useful string, as it does on Ubuntu,
        # we'll use the first string for the system type.
        self.useFixture(MockPatch("distro.name", return_value="Fooix"))
        self.useFixture(MockPatch("platform.system", return_value="FooOS"))
        self.useFixture(MockPatch("socket.gethostname", return_value="foo"))
        consumer = SystemWideConsumer("app name")
        self.assertEqual(consumer.key, "System-wide: Fooix (foo)")

    def test_empty_distro_name(self):
        # If distro.name returns an empty string, as it does on Windows and
        # Mac OS X, we fall back to the result of platform.system().
        self.useFixture(MockPatch("distro.name", return_value=""))
        self.useFixture(MockPatch("platform.system", return_value="BarOS"))
        self.useFixture(MockPatch("socket.gethostname", return_value="bar"))
        consumer = SystemWideConsumer("app name")
        self.assertEqual(consumer.key, "System-wide: BarOS (bar)")

    def test_broken_distro_name(self):
        # If distro.name raises an exception, we fall back to the result of
        # platform.system().
        self.useFixture(
            MockPatch("distro.name", side_effect=Exception("Oh noes!"))
        )
        self.useFixture(MockPatch("platform.system", return_value="BazOS"))
        self.useFixture(MockPatch("socket.gethostname", return_value="baz"))
        consumer = SystemWideConsumer("app name")
        self.assertEqual(consumer.key, "System-wide: BazOS (baz)")


class TestOAuthAuthorizer(TestCase):
    """Test for the OAuth Authorizer."""

    def test_save_to_and_load_from__path(self):
        # Credentials can be saved to and loaded from a file using
        # save_to_path() and load_from_path().
        temp_dir = self.useFixture(TempDir()).path
        credentials_path = os.path.join(temp_dir, "credentials")
        credentials = OAuthAuthorizer(
            "consumer.key",
            consumer_secret="consumer.secret",
            access_token=AccessToken("access.key", "access.secret"),
        )
        credentials.save_to_path(credentials_path)
        self.assertTrue(os.path.exists(credentials_path))

        # Make sure the file is readable and writable by the user, but
        # not by anyone else.
        self.assertEqual(
            stat.S_IMODE(os.stat(credentials_path).st_mode),
            stat.S_IREAD | stat.S_IWRITE,
        )

        loaded_credentials = OAuthAuthorizer.load_from_path(credentials_path)
        self.assertEqual(loaded_credentials.consumer.key, "consumer.key")
        self.assertEqual(loaded_credentials.consumer.secret, "consumer.secret")
        self.assertEqual(loaded_credentials.access_token.key, "access.key")
        self.assertEqual(
            loaded_credentials.access_token.secret, "access.secret"
        )


def test_suite():
    return unittest.TestLoader().loadTestsFromName(__name__)



# Copyright 2009 Canonical Ltd.

# This file is part of lazr.restfulclient.
#
# lazr.restfulclient is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, version 3 of the
# License.
#
# lazr.restfulclient is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with lazr.restfulclient. If not, see
# <http://www.gnu.org/licenses/>.

"""Classes for working with JSON."""

__metaclass__ = type
__all__ = ["DatetimeJSONEncoder"]

import datetime
from json import JSONEncoder


class DatetimeJSONEncoder(JSONEncoder):
    """A JSON encoder that understands datetime objects.

    Datetime objects are formatted according to ISO 1601.
    """

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return JSONEncoder.default(self, obj)



# Copyright 2008,2012 Canonical Ltd.

# This file is part of lazr.restfulclient.
#
# lazr.restfulclient is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# lazr.restfulclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with lazr.restfulclient.  If not, see
# <http://www.gnu.org/licenses/>.

"""Browser object to make requests of lazr.restful web services.

The `Browser` class does some massage of HTTP requests and responses,
and handles custom caches. It is not part of the public
lazr.restfulclient API. (But maybe it should be?)
"""

__metaclass__ = type
__all__ = [
    "Browser",
    "RestfulHttp",
    "ssl_certificate_validation_disabled",
]

import atexit
import errno
import os
import re
import shutil
import sys
import tempfile
from hashlib import md5
from io import BytesIO
from json import dumps

# Import sleep directly into the module so we can monkey-patch it
# during a test.
from time import sleep

from httplib2 import Http, urlnorm

try:
    from httplib2 import proxy_info_from_environment
except ImportError:
    from httplib2 import ProxyInfo

    proxy_info_from_environment = ProxyInfo.from_environment

try:
    # Python 3.
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from wadllib.application import Application

from lazr.restfulclient._json import DatetimeJSONEncoder
from lazr.restfulclient.errors import HTTPError, error_for
from lazr.uri import URI

if bytes is str:
    # Python 2
    unicode_type = unicode  # noqa: F821
    str_types = basestring  # noqa: F821
else:
    unicode_type = str
    str_types = str


# A drop-in replacement for httplib2's safename.  Substantially borrowed
# from httplib2, but its cache name format changed in 0.12.0 and we want to
# stick with the previous version.

re_url_scheme = re.compile(br"^\w+://")
re_url_scheme_s = re.compile(r"^\w+://")
re_slash = re.compile(br"[?/:|]+")


def safename(filename):
    """Return a filename suitable for the cache.

    Strips dangerous and common characters to create a filename we
    can use to store the cache in.
    """
    try:
        if isinstance(filename, bytes):
            filename_match = filename.decode("utf-8")
        else:
            filename_match = filename

        if re_url_scheme_s.match(filename_match):
            if isinstance(filename, bytes):
                filename = filename.decode("utf-8")
                filename = filename.encode("idna")
            else:
                filename = filename.encode("idna")
    except UnicodeError:
        pass
    if isinstance(filename, unicode_type):
        filename = filename.encode("utf-8")
    filemd5 = md5(filename).hexdigest()
    filename = re_url_scheme.sub(b"", filename)
    filename = re_slash.sub(b",", filename)

    # This is the part that we changed. In stock httplib2, the
    # filename is trimmed if it's longer than 200 characters, and then
    # a comma and a 32-character md5 sum are appended. This causes
    # problems on eCryptfs filesystems, where the maximum safe
    # filename length is closer to 143 characters.
    #
    # We take a (user-hackable) maximum filename length from
    # RestfulHttp and subtract 33 characters to make room for the comma
    # and the md5 sum.
    #
    # See:
    #  http://code.google.com/p/httplib2/issues/detail?id=92
    #  https://bugs.launchpad.net/bugs/344878
    #  https://bugs.launchpad.net/bugs/545197
    maximum_filename_length = RestfulHttp.maximum_cache_filename_length
    maximum_length_before_md5_sum = maximum_filename_length - 32 - 1
    if len(filename) > maximum_length_before_md5_sum:
        filename = filename[:maximum_length_before_md5_sum]
    return ",".join((filename.decode("utf-8"), filemd5))


def ssl_certificate_validation_disabled():
    """Whether the user has disabled SSL certificate connection.

    Some testing servers have broken certificates.  Rather than raising an
    error, we allow an environment variable,
    ``LP_DISABLE_SSL_CERTIFICATE_VALIDATION`` to disable the check.
    """
    return bool(os.environ.get("LP_DISABLE_SSL_CERTIFICATE_VALIDATION", False))


if os.path.exists("/etc/ssl/certs/ca-certificates.crt"):
    SYSTEM_CA_CERTS = "/etc/ssl/certs/ca-certificates.crt"
else:
    from httplib2 import CA_CERTS as SYSTEM_CA_CERTS


class RestfulHttp(Http):
    """An Http subclass with some custom behavior.

    This Http client uses the TE header instead of the Accept-Encoding
    header to ask for compressed representations. It also knows how to
    react when its cache is a MultipleRepresentationCache.
    """

    maximum_cache_filename_length = 143

    def __init__(
        self,
        authorizer=None,
        cache=None,
        timeout=None,
        proxy_info=proxy_info_from_environment,
    ):
        cert_disabled = ssl_certificate_validation_disabled()
        super(RestfulHttp, self).__init__(
            cache,
            timeout,
            proxy_info,
            disable_ssl_certificate_validation=cert_disabled,
            ca_certs=SYSTEM_CA_CERTS,
        )
        self.authorizer = authorizer
        if self.authorizer is not None:
            self.authorizer.authorizeSession(self)

    def _request(
        self,
        conn,
        host,
        absolute_uri,
        request_uri,
        method,
        body,
        headers,
        redirections,
        cachekey,
    ):
        """Use the authorizer to authorize an outgoing request."""
        if "authorization" in headers:
            # There's an authorization header left over from a
            # previous request that resulted in a redirect. Resources
            # protected by OAuth or HTTP Digest must send a distinct
            # Authorization header with each request, to prevent
            # playback attacks. Remove the Authorization header and
            # start again.
            del headers["authorization"]
        if self.authorizer is not None:
            self.authorizer.authorizeRequest(
                absolute_uri, method, body, headers
            )
        return super(RestfulHttp, self)._request(
            conn,
            host,
            absolute_uri,
            request_uri,
            method,
            body,
            headers,
            redirections,
            cachekey,
        )

    def _getCachedHeader(self, uri, header):
        """Retrieve a cached value for an HTTP header."""
        if isinstance(self.cache, MultipleRepresentationCache):
            return self.cache._getCachedHeader(uri, header)
        return None


class AtomicFileCache(object):
    """A FileCache that can be shared by multiple processes.

    Based on a patch found at
    <http://code.google.com/p/httplib2/issues/detail?id=125>.
    """

    TEMPFILE_PREFIX = ".temp"

    def __init__(self, cache, safe=safename):
        """Construct an ``AtomicFileCache``.

        :param cache: The directory to use as a cache.
        :param safe: A function that takes a key and returns a name that's
            safe to use as a filename.  The key must never return a string
            that begins with ``TEMPFILE_PREFIX``.  By default uses
            ``safename``.
        """
        self._cache_dir = os.path.normpath(cache)
        self._get_safe_name = safe
        try:
            os.makedirs(self._cache_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def _get_key_path(self, key):
        """Return the path on disk where ``key`` is stored."""
        safe_key = self._get_safe_name(key)
        if safe_key.startswith(self.TEMPFILE_PREFIX):
            # If the cache key starts with the tempfile prefix, then it's
            # possible that it will clash with a temporary file that we
            # create.
            raise ValueError(
                "Cache key cannot start with '%s'" % self.TEMPFILE_PREFIX
            )
        return os.path.join(self._cache_dir, safe_key)

    def get(self, key):
        """Get the value of ``key`` if set.

        This behaves slightly differently to ``FileCache`` in that if
        ``set()`` fails to store a key, this ``get()`` will behave as if that
        key were never set whereas ``FileCache`` returns the empty string.

        :param key: The key to retrieve.  Must be either bytes or unicode
            text.
        :return: The value of ``key`` if set, None otherwise.
        """
        cache_full_path = self._get_key_path(key)
        try:
            f = open(cache_full_path, "rb")
            try:
                return f.read()
            finally:
                f.close()
        except (IOError, OSError) as e:
            if e.errno != errno.ENOENT:
                raise

    def set(self, key, value):
        """Set ``key`` to ``value``.

        :param key: The key to set.  Must be either bytes or unicode text.
        :param value: The value to set ``key`` to.  Must be bytes.
        """
        # Open a temporary file
        handle, path_name = tempfile.mkstemp(
            prefix=self.TEMPFILE_PREFIX, dir=self._cache_dir
        )
        f = os.fdopen(handle, "wb")
        f.write(value)
        f.close()
        cache_full_path = self._get_key_path(key)
        # And rename atomically (on POSIX at least)
        if sys.platform == "win32" and os.path.exists(cache_full_path):
            os.unlink(cache_full_path)
        os.rename(path_name, cache_full_path)

    def delete(self, key):
        """Delete ``key`` from the cache.

        If ``key`` has not already been set then has no effect.

        :param key: The key to delete.  Must be either bytes or unicode text.
        """
        cache_full_path = self._get_key_path(key)
        try:
            os.remove(cache_full_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise


class MultipleRepresentationCache(AtomicFileCache):
    """A cache that can hold different representations of the same resource.

    If a resource has two representations with two media types,
    FileCache will only store the most recently fetched
    representation. This cache can keep track of multiple
    representations of the same resource.

    This class works on the assumption that outside calling code sets
    an instance's request_media_type attribute to the value of the
    'Accept' header before initiating the request.

    This class is very much not thread-safe, but FileCache isn't
    thread-safe anyway.
    """

    def __init__(self, cache):
        """Tell FileCache to call append_media_type when generating keys."""
        super(MultipleRepresentationCache, self).__init__(
            cache, self.append_media_type
        )
        self.request_media_type = None

    def append_media_type(self, key):
        """Append the request media type to the cache key.

        This ensures that representations of the same resource will be
        cached separately, so long as they're served as different
        media types.
        """
        if self.request_media_type is not None:
            key = key + "-" + self.request_media_type
        return safename(key)

    def _getCachedHeader(self, uri, header):
        """Retrieve a cached value for an HTTP header."""
        (scheme, authority, request_uri, cachekey) = urlnorm(uri)
        cached_value = self.get(cachekey)
        header_start = header + ":"
        if not isinstance(header_start, bytes):
            header_start = header_start.encode("utf-8")
        if cached_value is not None:
            for line in BytesIO(cached_value):
                if line.startswith(header_start):
                    return line[len(header_start) :].strip()
        return None


class Browser:
    """A class for making calls to lazr.restful web services."""

    NOT_MODIFIED = object()
    MAX_RETRIES = 6

    def __init__(
        self,
        service_root,
        credentials,
        cache=None,
        timeout=None,
        proxy_info=None,
        user_agent=None,
        max_retries=MAX_RETRIES,
    ):
        """Initialize, possibly creating a cache.

        If no cache is provided, a temporary directory will be used as
        a cache. The temporary directory will be automatically removed
        when the Python process exits.
        """
        if cache is None:
            cache = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, cache)
        if isinstance(cache, str_types):
            cache = MultipleRepresentationCache(cache)
        self._connection = service_root.httpFactory(
            credentials, cache, timeout, proxy_info
        )
        self.user_agent = user_agent
        self.max_retries = max_retries

    def _request_and_retry(self, url, method, body, headers):
        for retry_count in range(0, self.max_retries + 1):
            response, content = self._connection.request(
                url, method=method, body=body, headers=headers
            )
            if (
                response.status in [502, 503]
                and retry_count < self.max_retries
            ):
                # The server returned a 502 or 503. Sleep for 0, 1, 2,
                # 4, 8, 16, ... seconds and try again.
                sleep_for = int(2 ** (retry_count - 1))
                sleep(sleep_for)
            else:
                break
        # Either the request succeeded or we gave up.
        return response, content

    def _request(
        self,
        url,
        data=None,
        method="GET",
        media_type="application/json",
        extra_headers=None,
    ):
        """Create an authenticated request object."""
        # If the user is trying to get data that has been redacted,
        # give a helpful message.
        if url == "tag:launchpad.net:2008:redacted":
            raise ValueError(
                "You tried to access a resource that you "
                "don't have the server-side permission to see."
            )

        # Add extra headers for the request.
        headers = {"Accept": media_type}
        if self.user_agent is not None:
            headers["User-Agent"] = self.user_agent
        if isinstance(self._connection.cache, MultipleRepresentationCache):
            self._connection.cache.request_media_type = media_type
        if extra_headers is not None:
            headers.update(extra_headers)
        response, content = self._request_and_retry(
            str(url), method=method, body=data, headers=headers
        )
        if response.status == 304:
            # The resource didn't change.
            if content == b"":
                if (
                    "If-None-Match" in headers
                    or "If-Modified-Since" in headers
                ):
                    # The caller made a conditional request, and the
                    # condition failed. Rather than send an empty
                    # representation, which might be misinterpreted,
                    # send a special object that will let the calling code know
                    # that the resource was not modified.
                    return response, self.NOT_MODIFIED
                else:
                    # The caller didn't make a conditional request,
                    # but the response code is 304 and there's no
                    # content. The only way to handle this is to raise
                    # an error.
                    #
                    # We don't use error_for() here because 304 is not
                    # normally considered an error condition.
                    raise HTTPError(response, content)
            else:
                # XXX leonardr 2010/04/12 bug=httplib2#97
                #
                # Why is this check here? Why would there ever be any
                # content when the response code is 304? It's because of
                # an httplib2 bug that sometimes sets a 304 response
                # code when caching retrieved documents. When the
                # cached document is retrieved, we get a 304 response
                # code and a full representation.
                #
                # Since the cache lookup succeeded, the 'real'
                # response code is 200. This code undoes the bad
                # behavior in httplib2.
                response.status = 200
            return response, content
        # Turn non-2xx responses into appropriate HTTPError subclasses.
        error = error_for(response, content)
        if error is not None:
            raise error
        return response, content

    def get(self, resource_or_uri, headers=None, return_response=False):
        """GET a representation of the given resource or URI."""
        if isinstance(resource_or_uri, (str_types, URI)):
            url = resource_or_uri
        else:
            method = resource_or_uri.get_method("get")
            url = method.build_request_url()
        response, content = self._request(url, extra_headers=headers)
        if return_response:
            return (response, content)
        return content

    def get_wadl_application(self, url):
        """GET a WADL representation of the resource at the requested url."""
        wadl_type = "application/vnd.sun.wadl+xml"
        response, content = self._request(url, media_type=wadl_type)
        url = str(url)
        if not isinstance(content, bytes):
            content = content.encode("utf-8")
        return Application(url, content)

    def post(self, url, method_name, **kws):
        """POST a request to the web service."""
        kws["ws.op"] = method_name
        data = urlencode(kws)
        return self._request(url, data, "POST")

    def put(self, url, representation, media_type, headers=None):
        """PUT the given representation to the URL."""
        extra_headers = {"Content-Type": media_type}
        if headers is not None:
            extra_headers.update(headers)
        return self._request(
            url, representation, "PUT", extra_headers=extra_headers
        )

    def delete(self, url):
        """DELETE the resource at the given URL."""
        self._request(url, method="DELETE")
        return None

    def patch(self, url, representation, headers=None):
        """PATCH the object at url with the updated representation."""
        extra_headers = {"Content-Type": "application/json"}
        if headers is not None:
            extra_headers.update(headers)
        # httplib2 doesn't know about the PATCH method, so we need to
        # do some work ourselves. Pull any cached value of "ETag" out
        # and use it as the value for "If-Match".
        cached_etag = self._connection._getCachedHeader(str(url), "etag")
        if cached_etag is not None and not self._connection.ignore_etag:
            # http://www.w3.org/1999/04/Editing/
            headers["If-Match"] = cached_etag

        return self._request(
            url,
            dumps(representation, cls=DatetimeJSONEncoder),
            "PATCH",
            extra_headers=extra_headers,
        )



# Copyright 2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.restfulclient
#
# lazr.restfulclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.restfulclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.restfulclient.  If not, see <http://www.gnu.org/licenses/>.
"Test harness for doctests."

# pylint: disable-msg=E0611,W0142

__metaclass__ = type
__all__ = [
    "load_tests",
]

import atexit
import doctest
import os

import wsgi_intercept
from pkg_resources import (
    cleanup_resources,
    resource_exists,
    resource_filename,
    resource_listdir,
)
from wsgi_intercept.httplib2_intercept import install, uninstall

# We avoid importing anything from lazr.restful into the module level,
# so that standalone_tests() can run without any support from
# lazr.restful.

DOCTEST_FLAGS = (
    doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF
)


def setUp(test):
    from lazr.restful.example.base.tests.test_integration import WSGILayer

    install()
    wsgi_intercept.add_wsgi_intercept(
        "cookbooks.dev", 80, WSGILayer.make_application
    )


def tearDown(test):
    from zope.component import getUtility

    from lazr.restful.example.base.interfaces import IFileManager

    uninstall()
    file_manager = getUtility(IFileManager)
    file_manager.files = {}
    file_manager.counter = 0


def find_doctests(suffix, ignore_suffix=None):
    """Find doctests matching a certain suffix."""
    doctest_files = []
    # Match doctests against the suffix.
    if resource_exists("lazr.restfulclient", "docs"):
        for name in resource_listdir("lazr.restfulclient", "docs"):
            if ignore_suffix is not None and name.endswith(ignore_suffix):
                continue
            if name.endswith(suffix):
                doctest_files.append(
                    os.path.abspath(
                        resource_filename(
                            "lazr.restfulclient", "docs/%s" % name
                        )
                    )
                )
    return doctest_files


def load_tests(loader, tests, pattern):
    """Load all the doctests."""
    from lazr.restful.example.base.tests.test_integration import WSGILayer

    atexit.register(cleanup_resources)
    restful_suite = doctest.DocFileSuite(
        *find_doctests(".rst", ignore_suffix=".standalone.rst"),
        module_relative=False,
        optionflags=DOCTEST_FLAGS,
        setUp=setUp,
        tearDown=tearDown
    )
    restful_suite.layer = WSGILayer
    tests.addTest(restful_suite)
    tests.addTest(
        doctest.DocFileSuite(
            *find_doctests(".standalone.rst"),
            module_relative=False,
            optionflags=DOCTEST_FLAGS
        )
    )
    return tests



# Copyright 2012 Canonical Ltd.

# This file is part of lazr.restfulclient.
#
# lazr.restfulclient is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, version 3 of the
# License.
#
# lazr.restfulclient is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.restfulclient. If not, see <http://www.gnu.org/licenses/>.

"""Tests for the atomic file cache."""

__metaclass__ = type

import shutil
import sys
import tempfile
import unittest

import httplib2

from lazr.restfulclient._browser import AtomicFileCache, safename

PY3 = sys.version_info[0] >= 3
if PY3:
    binary_type = bytes
else:
    binary_type = str


class TestFileCacheInterface(unittest.TestCase):
    """Tests for ``AtomicFileCache``."""

    file_cache_factory = httplib2.FileCache

    unicode_bytes = b"pa\xc9\xaa\xce\xb8\xc9\x99n"
    unicode_text = unicode_bytes.decode("utf-8")

    def setUp(self):
        super(TestFileCacheInterface, self).setUp()
        self.cache_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.cache_dir)
        super(TestFileCacheInterface, self).tearDown()

    def make_file_cache(self):
        """Make a FileCache-like object to be tested."""
        return self.file_cache_factory(self.cache_dir, safename)

    def test_get_non_existent_key(self):
        # get() returns None if the key does not exist.
        cache = self.make_file_cache()
        self.assertIs(None, cache.get("nonexistent"))

    def test_set_key(self):
        # A key set with set() can be got by get().
        cache = self.make_file_cache()
        cache.set("key", b"value")
        self.assertEqual(b"value", cache.get("key"))

    def test_set_twice_overrides(self):
        # Setting a key again overrides the value.
        cache = self.make_file_cache()
        cache.set("key", b"value")
        cache.set("key", b"new-value")
        self.assertEqual(b"new-value", cache.get("key"))

    def test_delete_absent_key(self):
        # Deleting a key that's not there does nothing.
        cache = self.make_file_cache()
        cache.delete("nonexistent")
        self.assertIs(None, cache.get("nonexistent"))

    def test_delete_key(self):
        # A key once set can be deleted.  Further attempts to get that key
        # return None.
        cache = self.make_file_cache()
        cache.set("key", b"value")
        cache.delete("key")
        self.assertIs(None, cache.get("key"))

    def test_get_non_string_key(self):
        # get() raises TypeError if asked to get a non-string key.
        cache = self.make_file_cache()
        self.assertRaises(TypeError, cache.get, 42)

    def test_delete_non_string_key(self):
        # delete() raises TypeError if asked to delete a non-string key.
        cache = self.make_file_cache()
        self.assertRaises(TypeError, cache.delete, 42)

    def test_set_non_string_key(self):
        # set() raises TypeError if asked to set a non-string key.
        cache = self.make_file_cache()
        self.assertRaises(TypeError, cache.set, 42, "the answer")

    def test_set_non_string_value(self):
        # set() raises TypeError if asked to set a key to a non-string value.
        # Attempts to retrieve that value return the empty string.  This is
        # probably a bug in httplib2.FileCache.
        cache = self.make_file_cache()
        self.assertRaises(TypeError, cache.set, "answer", 42)
        self.assertEqual(b"", cache.get("answer"))

    def test_get_unicode(self):
        # get() can retrieve unicode keys.
        cache = self.make_file_cache()
        self.assertIs(None, cache.get(self.unicode_text))

    def test_set_unicode_keys(self):
        cache = self.make_file_cache()
        cache.set(self.unicode_text, b"value")
        self.assertEqual(b"value", cache.get(self.unicode_text))

    def test_set_unicode_value(self):
        # set() cannot store unicode values.  Values must be bytes.
        cache = self.make_file_cache()
        error = TypeError if PY3 else UnicodeEncodeError
        self.assertRaises(error, cache.set, "key", self.unicode_text)

    def test_delete_unicode(self):
        # delete() can remove unicode keys.
        cache = self.make_file_cache()
        cache.set(self.unicode_text, b"value")
        cache.delete(self.unicode_text)
        self.assertIs(None, cache.get(self.unicode_text))


class TestAtomicFileCache(TestFileCacheInterface):
    """Tests for ``AtomicFileCache``."""

    file_cache_factory = AtomicFileCache

    @staticmethod
    def prefix_safename(x):
        if isinstance(x, binary_type):
            x = x.decode("utf-8")
        return AtomicFileCache.TEMPFILE_PREFIX + x

    def test_set_non_string_value(self):
        # set() raises TypeError if asked to set a key to a non-string value.
        # Attempts to retrieve that value act is if it were never set.
        #
        # Note: This behaviour differs from httplib2.FileCache.
        cache = self.make_file_cache()
        self.assertRaises(TypeError, cache.set, "answer", 42)
        self.assertIs(None, cache.get("answer"))

    # Implementation-specific tests follow.

    def test_bad_safename_get(self):
        safename = self.prefix_safename
        cache = AtomicFileCache(self.cache_dir, safename)
        self.assertRaises(ValueError, cache.get, "key")

    def test_bad_safename_set(self):
        safename = self.prefix_safename
        cache = AtomicFileCache(self.cache_dir, safename)
        self.assertRaises(ValueError, cache.set, "key", b"value")

    def test_bad_safename_delete(self):
        safename = self.prefix_safename
        cache = AtomicFileCache(self.cache_dir, safename)
        self.assertRaises(ValueError, cache.delete, "key")



#!/usr/bin/env python

# Copyright 2009-2018 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.restfulclient
#
# lazr.restfulclient is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.restfulclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.restfulclient.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import find_packages, setup


# generic helpers primarily for the long_description
def generate(*docname_or_string):
    marker = ".. pypi description ends here"
    res = []
    for value in docname_or_string:
        if value.endswith(".rst"):
            with open(value) as f:
                value = f.read()
            idx = value.find(marker)
            if idx >= 0:
                value = value[:idx]
        res.append(value)
        if not value.endswith("\n"):
            res.append("")
    return "\n".join(res)


# end generic helpers


tests_require = [
    "fixtures>=1.3.0",
    "lazr.authentication",
    "lazr.restful>=0.11.0",
    'mock; python_version < "3"',
    "oauth",
    "testtools",
    "wsgi_intercept",
    "zope.testrunner",
]

setup(
    name="lazr.restfulclient",
    version="0.14.6",
    namespace_packages=["lazr"],
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    maintainer="LAZR Developers",
    maintainer_email="lazr-developers@lists.launchpad.net",
    description=open("README.rst").readline().strip(),
    long_description=generate(
        "src/lazr/restfulclient/docs/index.rst", "NEWS.rst"
    ),
    license="LGPL v3",
    install_requires=[
        "distro",
        'httplib2; python_version < "3"',
        'httplib2>=0.7.7; python_version >= "3"',
        'importlib-metadata; python_version < "3.8"',
        "oauthlib",
        "setuptools",
        "six",
        "wadllib>=1.1.4",
    ],
    url="https://launchpad.net/lazr.restfulclient",
    project_urls={
        "Source": "https://code.launchpad.net/lazr.restfulclient",
        "Issue Tracker": "https://bugs.launchpad.net/lazr.restfulclient",
        "Documentation": "https://lazrrestfulclient.readthedocs.io/en/latest/",
    },
    download_url="https://launchpad.net/lazr.restfulclient/+download",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",  # noqa: E501
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    tests_require=tests_require,
    extras_require=dict(
        docs=["Sphinx"],
        test=tests_require,
    ),
    test_suite="lazr.restfulclient.tests",
)



# Copyright 2009-2018 Canonical Ltd.

# This file is part of lazr.restfulclient.
#
# lazr.restfulclient is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# lazr.restfulclient is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with lazr.restfulclient.  If not, see
# <http://www.gnu.org/licenses/>.

"""OAuth classes for use with lazr.restfulclient."""


try:
    # Python 3, SafeConfigParser was renamed to just ConfigParser.
    from configparser import ConfigParser as SafeConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser

import os
import platform
import socket
import stat

import six
from oauthlib import oauth1
from six.moves.urllib.parse import parse_qs, urlencode

from lazr.restfulclient.authorize import HttpAuthorizer
from lazr.restfulclient.errors import CredentialsFileError

__metaclass__ = type
__all__ = [
    "AccessToken",
    "Consumer",
    "OAuthAuthorizer",
    "SystemWideConsumer",
]


CREDENTIALS_FILE_VERSION = "1"


# For compatibility, Consumer and AccessToken are defined using terminology
# from the older oauth library rather than the newer oauthlib.
class Consumer:
    """An OAuth consumer (application)."""

    def __init__(self, key, secret="", application_name=None):
        """Initialize

        :param key: The OAuth consumer key
        :param secret: The OAuth consumer secret. Don't use this. It's
            a misfeature, and lazr.restful doesn't expect it.
        :param application_name: An application name, if different
            from the consumer key. If present, this will be used in
            the User-Agent header.
        """
        self.key = key
        self.secret = secret
        self.application_name = application_name


class AccessToken:
    """An OAuth access token."""

    def __init__(self, key, secret="", context=None):
        self.key = key
        self.secret = secret
        self.context = context

    def to_string(self):
        return urlencode(
            [
                ("oauth_token_secret", self.secret),
                ("oauth_token", self.key),
            ]
        )

    __str__ = to_string

    @classmethod
    def from_string(cls, s):
        params = parse_qs(s, keep_blank_values=False)
        key = params["oauth_token"][0]
        secret = params["oauth_token_secret"][0]
        return cls(key, secret)


class TruthyString(six.text_type):
    """A Unicode string which is always true."""

    def __bool__(self):
        return True

    __nonzero__ = __bool__


class SystemWideConsumer(Consumer):
    """A consumer associated with the logged-in user rather than an app.

    This can be used to share a single OAuth token among multiple
    desktop applications. The OAuth consumer key will be derived from
    system information (platform and hostname).
    """

    KEY_FORMAT = "System-wide: %s (%s)"

    def __init__(self, application_name, secret=""):
        """Constructor.

        :param application_name: An application name. This will be
            used in the User-Agent header.
        :param secret: The OAuth consumer secret. Don't use this. It's
            a misfeature, and lazr.restful doesn't expect it.
        """
        super(SystemWideConsumer, self).__init__(
            self.consumer_key, secret, application_name
        )

    @property
    def consumer_key(self):
        """The system-wide OAuth consumer key for this computer.

        This key identifies the platform and the computer's
        hostname. It does not identify the active user.
        """
        try:
            import distro

            distname = distro.name()
        except Exception:
            # This can happen due to various kinds of failures with the data
            # sources used by the distro module.
            distname = ""
        if distname == "":
            distname = platform.system()  # (eg. "Windows")
        return self.KEY_FORMAT % (distname, socket.gethostname())


class OAuthAuthorizer(HttpAuthorizer):
    """A client that signs every outgoing request with OAuth credentials."""

    def __init__(
        self,
        consumer_name=None,
        consumer_secret="",
        access_token=None,
        oauth_realm="OAuth",
        application_name=None,
    ):
        self.consumer = None
        if consumer_name is not None:
            self.consumer = Consumer(
                consumer_name, consumer_secret, application_name
            )
        self.access_token = access_token
        self.oauth_realm = oauth_realm

    @property
    def user_agent_params(self):
        """Any information necessary to identify this user agent.

        In this case, the OAuth consumer name.
        """
        params = {}
        if self.consumer is None:
            return params
        params["oauth_consumer"] = self.consumer.key
        if self.consumer.application_name is not None:
            params["application"] = self.consumer.application_name
        return params

    def load(self, readable_file):
        """Load credentials from a file-like object.

        This overrides the consumer and access token given in the constructor
        and replaces them with the values read from the file.

        :param readable_file: A file-like object to read the credentials from
        :type readable_file: Any object supporting the file-like `read()`
            method
        """
        # Attempt to load the access token from the file.
        parser = SafeConfigParser()
        if hasattr(parser, "read_file"):
            reader = parser.read_file
        else:
            reader = parser.readfp
        reader(readable_file)
        # Check the version number and extract the access token and
        # secret.  Then convert these to the appropriate instances.
        if not parser.has_section(CREDENTIALS_FILE_VERSION):
            raise CredentialsFileError(
                "No configuration for version %s" % CREDENTIALS_FILE_VERSION
            )
        consumer_key = parser.get(CREDENTIALS_FILE_VERSION, "consumer_key")
        consumer_secret = parser.get(
            CREDENTIALS_FILE_VERSION, "consumer_secret"
        )
        self.consumer = Consumer(consumer_key, consumer_secret)
        access_token = parser.get(CREDENTIALS_FILE_VERSION, "access_token")
        access_secret = parser.get(CREDENTIALS_FILE_VERSION, "access_secret")
        self.access_token = AccessToken(access_token, access_secret)

    @classmethod
    def load_from_path(cls, path):
        """Convenience method for loading credentials from a file.

        Open the file, create the Credentials and load from the file,
        and finally close the file and return the newly created
        Credentials instance.

        :param path: In which file the credential file should be saved.
        :type path: string
        :return: The loaded Credentials instance.
        :rtype: `Credentials`
        """
        credentials = cls()
        credentials_file = open(path, "r")
        credentials.load(credentials_file)
        credentials_file.close()
        return credentials

    def save(self, writable_file):
        """Write the credentials to the file-like object.

        :param writable_file: A file-like object to write the credentials to
        :type writable_file: Any object supporting the file-like `write()`
            method
        :raise CredentialsFileError: when there is either no consumer or no
            access token
        """
        if self.consumer is None:
            raise CredentialsFileError("No consumer")
        if self.access_token is None:
            raise CredentialsFileError("No access token")

        parser = SafeConfigParser()
        parser.add_section(CREDENTIALS_FILE_VERSION)
        parser.set(CREDENTIALS_FILE_VERSION, "consumer_key", self.consumer.key)
        parser.set(
            CREDENTIALS_FILE_VERSION, "consumer_secret", self.consumer.secret
        )
        parser.set(
            CREDENTIALS_FILE_VERSION, "access_token", self.access_token.key
        )
        parser.set(
            CREDENTIALS_FILE_VERSION, "access_secret", self.access_token.secret
        )
        parser.write(writable_file)

    def save_to_path(self, path):
        """Convenience method for saving credentials to a file.

        Create the file, call self.save(), and close the
        file. Existing files are overwritten. The resulting file will
        be readable and writable only by the user.

        :param path: In which file the credential file should be saved.
        :type path: string
        """
        credentials_file = os.fdopen(
            os.open(
                path,
                (os.O_CREAT | os.O_TRUNC | os.O_WRONLY),
                (stat.S_IREAD | stat.S_IWRITE),
            ),
            "w",
        )
        self.save(credentials_file)
        credentials_file.close()

    def authorizeRequest(self, absolute_uri, method, body, headers):
        """Sign a request with OAuth credentials."""
        client = oauth1.Client(
            self.consumer.key,
            client_secret=self.consumer.secret,
            resource_owner_key=TruthyString(self.access_token.key or ""),
            resource_owner_secret=self.access_token.secret,
            signature_method=oauth1.SIGNATURE_PLAINTEXT,
            realm=self.oauth_realm,
        )
        # The older oauth library (which may still be used on the server)
        # requires the oauth_token parameter to be present and will fail
        # authentication if it isn't.  This hack forces it to be present
        # even if its value is the empty string.
        client.resource_owner_key = TruthyString(client.resource_owner_key)
        _, signed_headers, _ = client.sign(absolute_uri)
        for key, value in signed_headers.items():
            # client.sign returns Unicode headers; convert these to native
            # strings.
            if six.PY2:
                key = key.encode("UTF-8")
                value = value.encode("UTF-8")
            headers[key] = value
